import warnings
import itertools
import pytest
import torch
from xitorch._utils.unique import Uniquifier
from xitorch._utils.decorators import deprecated
from xitorch._utils.tensor import create_random_square_matrix, \
    create_random_ortho_matrix

@deprecated("06 Oct 2020")
def deprfunc():
    return 3

@deprecated("06 Oct 2020")
class DeprClass():
    def __init__(self):
        self.a = 3

class DeprMethod():
    def __init__(self):
        self.a = 4

    @deprecated("06 Oct 2020")
    def method(self):
        return self.a + 2

def test_uniquifier():
    obj1 = [1]
    obj2 = [2]
    obj3 = [3]
    obj4 = [4]
    obj5 = [5]
    objs = [obj1, obj2, obj1, obj2, obj3, obj3]
    objsA = [obj2, obj1, obj2, obj1, obj4, obj4]

    uniq = Uniquifier(objs)
    unique_objs = uniq.get_unique_objs()
    assert len(unique_objs) == 3
    assert unique_objs[0] is obj1
    assert unique_objs[1] is obj2
    assert unique_objs[2] is obj3

    unique_objsA = uniq.get_unique_objs(objsA)
    assert len(unique_objsA) == 3
    assert unique_objsA[0] is obj2
    assert unique_objsA[1] is obj1
    assert unique_objsA[2] is obj4

    unique_objs2 = [obj3, obj4, obj5]
    objs2 = uniq.map_unique_objs(unique_objs2)
    assert len(objs2) == len(objs)
    assert objs2[0] is obj3
    assert objs2[1] is obj4
    assert objs2[2] is obj3
    assert objs2[3] is obj4
    assert objs2[4] is obj5
    assert objs2[5] is obj5

def test_uniquifier_error():
    obj1 = [1]
    obj2 = [2]
    obj3 = [3]
    obj4 = [4]
    objs = [obj1, obj2, obj1, obj1, obj2]
    objs2 = [obj1, obj2, obj1, obj1, obj2, obj2]

    uniq = Uniquifier(objs)
    try:
        unique_objs = uniq.get_unique_objs(objs2)
        assert False, "Expected a RuntimeError"
    except RuntimeError:
        pass

    try:
        objs3 = uniq.map_unique_objs(objs)
        assert False, "Expected a RuntimeError"
    except RuntimeError:
        pass

def test_deprecated_func():
    with warnings.catch_warnings(record=True) as w:
        a = deprfunc()
        assert len(w) == 1
        msg = str(w[0].message)
        assert "deprfunc" in msg
        assert "deprecated" in msg
        assert "06 Oct 2020" in msg

def test_deprecated_class():
    with warnings.catch_warnings(record=True) as w:
        a = DeprClass()
        assert len(w) == 1
        msg = str(w[0].message)
        assert "DeprClass" in msg
        assert "deprecated" in msg
        assert "06 Oct 2020" in msg

def test_deprecated_method():
    with warnings.catch_warnings(record=True) as w:
        a = DeprMethod()
        b = a.method()
        assert len(w) == 1
        msg = str(w[0].message)
        assert "DeprMethod.method" in msg
        assert "deprecated" in msg
        assert "06 Oct 2020" in msg

@pytest.mark.parametrize(
    "is_hermitian,minmaxeival",
    itertools.product([False, True], [(-1.0, 1.0), (0.0, 1.0), (0.5, 1.0)])
)
def test_create_random_matrix(is_hermitian, minmaxeival):
    n = 100
    min_eival, max_eival = minmaxeival
    a = create_random_square_matrix(
        n, is_hermitian=is_hermitian,
        min_eival=min_eival, max_eival=max_eival)
    assert a.shape[0] == n
    assert a.shape[1] == n
    assert a.dtype == torch.float64
    dtype = a.dtype
    if is_hermitian:
        assert torch.allclose(a, a.transpose(-2, -1))
        eivals = torch.linalg.eigvalsh(a)
    else:
        eivals = torch.linalg.eigvals(a)
        assert torch.allclose(torch.imag(eivals), torch.imag(eivals) * 0, atol=1e-4)
        eivals = torch.real(eivals)
    assert torch.allclose(eivals.min(), torch.tensor(min_eival, dtype=dtype),
                          atol=1e-4)
    assert torch.allclose(eivals.max(), torch.tensor(max_eival, dtype=dtype),
                          atol=1e-4)

def test_create_random_ortho_matrix():
    n = 100
    a = create_random_ortho_matrix(n)
    eye = torch.eye(n, dtype=a.dtype)
    assert torch.allclose(torch.matmul(a.transpose(-2, -1), a), eye)



import torch
import pytest
from typing import List
from xitorch._core.editable_module import EditableModule
from xitorch._core.pure_function import get_pure_function
from xitorch._utils.exceptions import GetSetParamsError

##############
# test the assertion with methods with various problems
class ModuleTest(EditableModule):
    def __init__(self, a: torch.Tensor) -> None:
        self.a = a
        self.c = a * 2.
        self.d = a * 3.
        self.e = a + 3.
        self.fint = 1
        self.g = a + 1.
        self.aa = a
        self.aaa = a
        self.dctparams = {
            0: a + 0.1,
            1: a + 1.,
            2: a + 2.,
        }
        self.listparams = [a + 0.12, a + 2., a + 4]

    def method_no_preserve1(self, b: torch.Tensor) -> torch.Tensor:
        # this method changes a parameter
        self.a += b
        return self.a

    def method_no_preserve2(self, b: torch.Tensor) -> torch.Tensor:
        # this method adds a parameter to the object
        self.b = b
        return self.b * 2.0

    def method_dict_correct(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self.dctparams[0] + self.dctparams[2]

    def method_dict_missing(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self.dctparams[0] + self.dctparams[2]

    def method_dict_excess(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self.dctparams[0] + self.dctparams[2]

    def method_list_correct(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self.listparams[0] + self.listparams[2]

    def method_list_missing(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self.listparams[0] + self.listparams[2]

    def method_list_excess(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self.listparams[0] + self.listparams[2]

    def method_duplicate_missing(self, b: torch.Tensor) -> torch.Tensor:
        # `aa` is `a`, but `aa` will be missing in getparamnames
        return self._dummy_fcn(b) + self.aa

    def method_duplicate_excess(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self.aa

    def method_duplicate_correct(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self.aa

    def method_correct_getsetparams(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b)

    def method_correct_getsetparams2(self, b: torch.Tensor, b2: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b) + self._dummy_fcn(b2)

    def method_nontensor_getparams(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b)

    def method_missing_getparams(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b)

    def method_excess_getparams(self, b: torch.Tensor) -> torch.Tensor:
        return self._dummy_fcn(b)

    def _dummy_fcn(self, b: torch.Tensor) -> torch.Tensor:
        return self.a + b + self.c + self.d + self.e + self.fint

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        if methodname == "method_no_preserve1":
            return [prefix + "a"]
        elif methodname == "method_no_preserve1":
            return []

        elif methodname == "method_dict_correct":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "dctparams[0]", prefix + "dctparams[2]"]
        elif methodname == "method_dict_missing":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "dctparams[0]"]
        elif methodname == "method_dict_excess":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "dctparams[0]", prefix + "dctparams[1]",
                    prefix + "dctparams[2]"]

        elif methodname == "method_list_correct":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "listparams[0]", prefix + "listparams[2]"]
        elif methodname == "method_list_missing":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "listparams[0]"]
        elif methodname == "method_list_excess":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "listparams[0]", prefix + "listparams[1]",
                    prefix + "listparams[2]"]

        elif methodname == "method_duplicate_correct":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "aa"]
        elif methodname == "method_duplicate_missing":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e"]
        elif methodname == "method_duplicate_excess":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "aa", prefix + "aaa"]

        elif methodname == "method_correct_getsetparams":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e"]
        elif methodname == "method_correct_getsetparams2":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e"]

        elif methodname == "method_nontensor_getparams":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "fint"]
        elif methodname == "method_missing_getparams":
            return [prefix + "a", prefix + "c", prefix + "d"]
        elif methodname == "method_excess_getparams":
            return [prefix + "a", prefix + "c", prefix + "d", prefix + "e",
                    prefix + "g"]
        else:
            raise KeyError("getparams for %s is not implemented" % methodname)


a = torch.tensor([1.])
b = torch.tensor([2.1])
model = ModuleTest(a)

@pytest.mark.filterwarnings("error")
def test_correct():
    correct_methods = {
        "method_correct_getsetparams": (b,),
        "method_correct_getsetparams2": (b, b),
        "method_duplicate_correct": (b,),
        "method_duplicate_missing": (b,),
        "method_duplicate_excess": (b,),
        "method_dict_correct": (b,),
        "method_list_correct": (b,),
    }
    for m in correct_methods:
        model.assertparams(getattr(model, m), *correct_methods[m])

def test_error_getsetparams():
    error_methods = [
        "method_no_preserve1",
        "method_no_preserve2",
        "method_nontensor_getparams",
    ]
    for methodname in error_methods:
        try:
            print(methodname)
            model.assertparams(getattr(model, methodname), b)
            assert False, "A GetSetParamsError must be raised in this case"
        except GetSetParamsError:
            pass

def test_warning_getsetparams():
    warning_methods = [
        "method_missing_getparams",
        "method_excess_getparams",
        "method_dict_missing",
        "method_dict_excess",
        "method_list_missing",
        "method_list_excess",
    ]
    for methodname in warning_methods:
        with pytest.warns(UserWarning):
            model.assertparams(getattr(model, methodname), b)

def test_get_unique_params_leaves():
    # test getuniqueparams in EditableModule

    # redefine the model in this case only
    a = torch.tensor([1.], requires_grad=True)
    model = ModuleTest(a)

    params = model.getuniqueparams(methodname="method_list_correct", onlyleaves=True)
    assert len(params) == 1
    assert params[0] is a

    params = model.getuniqueparams(methodname="method_list_correct")
    assert len(params) == 6

##############
# test the wrap function to make it a functional
def test_edit_simple():
    pfcn = get_pure_function(model.method_correct_getsetparams)
    objparams = pfcn.objparams()
    assert len(objparams) == 4
    newb = torch.tensor([1.])
    newobjparams = [torch.tensor(1.0 * i + 2) for i in range(len(objparams))]
    with pfcn.useobjparams(newobjparams):
        f = pfcn(newb)
    assert torch.allclose(f, f * 0 + 16)  # (1+2+3+4+5+1)

    pfcn2 = get_pure_function(model.method_correct_getsetparams2)
    objparams2 = pfcn2.objparams()
    assert len(objparams2) == 4
    newparams2 = [torch.tensor(1.0 * i + 1) for i in range(2)]
    newobjparams2 = [torch.tensor(1.0 * i + 3) for i in range(len(objparams2))]
    with pfcn2.useobjparams(newobjparams2):
        f2 = pfcn2(*newparams2)
    assert torch.allclose(f2, f2 * 0 + 41)  # (1+3+4+5+6+1) + (2+3+4+5+6+1)

def test_edit_duplicate():
    pfcn = get_pure_function(model.method_duplicate_correct)
    objparams = pfcn.objparams()
    assert len(objparams) == 4
    assert objparams[0] is model.a
    assert objparams[0] is model.aa  # aa is a duplicate of a
    newparams = [torch.tensor(1.0 * i + 1) for i in range(1)]
    newobjparams = [torch.tensor(1.0 * i + 2) for i in range(len(objparams))]
    with pfcn.useobjparams(newobjparams):
        f = pfcn(*newparams)
    assert torch.allclose(f, f * 0 + 18)  # (1+2+3+4+5+1) + 2

def test_edit_dict():
    pfcn = get_pure_function(model.method_dict_correct)
    objparams = pfcn.objparams()
    assert len(objparams) == 6
    assert objparams[-1] is model.dctparams[2]
    assert objparams[-2] is model.dctparams[0]
    newparams = [torch.tensor(1.0 * i + 1) for i in range(1)]
    newobjparams = [torch.tensor(1.0 * i + 2) for i in range(len(objparams))]
    with pfcn.useobjparams(newobjparams):
        f = pfcn(*newparams)
    assert torch.allclose(f, f * 0 + 29)  # (1+2+3+4+5+1) + 6+7

def test_edit_list():
    pfcn = get_pure_function(model.method_list_correct)
    objparams = pfcn.objparams()
    assert len(objparams) == 6
    assert objparams[-1] is model.listparams[2]
    assert objparams[-2] is model.listparams[0]
    newparams = [torch.tensor(1.0 * i + 1) for i in range(1)]
    newobjparams = [torch.tensor(1.0 * i + 2) for i in range(len(objparams))]
    with pfcn.useobjparams(newobjparams):
        f = pfcn(*newparams)
    assert torch.allclose(f, f * 0 + 29)  # (1+2+3+4+5+1) + 6+7



import xitorch
import torch
import pytest
from xitorch._core.pure_function import get_pure_function, PureFunction

def func1(x, a, b):
    return x * a + b

@torch.jit.script
def jitfunc1(x, a, b):
    return x * a + b

class TorchModule(torch.nn.Module):
    def __init__(self, a, b):
        super().__init__()
        self.a = a
        self.b = b

    def forward(self, x):
        return x * self.a + self.b

class MyModule(xitorch.EditableModule):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def forward(self, x):
        return x * self.a + self.b

    def forward2(self, x):
        return x + self.a

    def getparamnames(self, methodname, prefix=""):
        if methodname == "forward":
            return [prefix + "a", prefix + "b"]
        elif methodname == "forward2":
            return [prefix + "a"]
        else:
            raise KeyError()

@pytest.mark.parametrize("fcn", [func1, jitfunc1])
def test_pure_function(fcn):
    pfunc1 = get_pure_function(fcn)
    assert isinstance(pfunc1, PureFunction)
    assert len(pfunc1.objparams()) == 0
    a = torch.tensor(2.0)
    b = torch.tensor(1.0)
    x = torch.tensor(1.5)
    res = x * a + b

    expr = lambda x, a, b: x * a + b
    runtest_pfunc(pfunc1, (x, a, b), expr)

@pytest.mark.parametrize("clss", [TorchModule, MyModule])
def test_module_pfunc(clss):
    a = torch.nn.Parameter(torch.tensor(2.0))
    b = torch.nn.Parameter(torch.tensor(1.0))
    x = torch.tensor(1.5)
    module = clss(a, b)
    pfunc = get_pure_function(module.forward)

    expr = lambda x, a, b: x * a + b
    runtest_pfunc(pfunc, (x,), expr)

@pytest.mark.parametrize("fcn", [func1, jitfunc1])
def test_make_sibling_fcn(fcn):
    a = torch.tensor(2.0)
    b = torch.tensor(1.0)
    x = torch.tensor(1.5)

    @xitorch.make_sibling(fcn)
    def fminusx(x, a, b):
        return fcn(x, a, b) - x

    assert isinstance(fminusx, PureFunction)
    assert len(fminusx.objparams()) == 0

    expr = lambda x, a, b: x * a + b - x
    runtest_pfunc(fminusx, (x, a, b), expr)

@pytest.mark.parametrize("clss", [TorchModule, MyModule])
def test_make_sibling_method(clss):
    a = torch.nn.Parameter(torch.tensor(2.0))
    b = torch.nn.Parameter(torch.tensor(1.0))
    x = torch.tensor(1.5)
    module = clss(a, b)

    @xitorch.make_sibling(module.forward)
    def fminusx(x):
        return module.forward(x) - x

    assert isinstance(fminusx, PureFunction)
    assert len(fminusx.objparams()) == 2

    expr = lambda x, a, b: x * a + b - x
    runtest_pfunc(fminusx, (x,), expr)

def test_make_sibling_multiple():
    a = torch.nn.Parameter(torch.tensor(2.0))
    b = torch.nn.Parameter(torch.tensor(1.0))
    x = torch.tensor(1.5)
    module1 = TorchModule(a, b)
    module2 = MyModule(a, b * 2)

    @xitorch.make_sibling(module1.forward, module2.forward)
    def newfcn(x):
        return module1.forward(x) + module2.forward(x) * 2

    assert isinstance(newfcn, PureFunction)
    assert len(newfcn.objparams()) == 3  # not 4, because a is identical

    expr = lambda x, a, b, b2: (x * a + b) + (x * a + b2) * 2
    runtest_pfunc(newfcn, (x,), expr)

def test_make_sibling_multiple_nonunique_objs():
    a = torch.nn.Parameter(torch.tensor(2.0))
    b = torch.nn.Parameter(torch.tensor(1.0))
    x = torch.tensor(1.5)
    module = MyModule(a, b)

    @xitorch.make_sibling(module.forward, module.forward2)
    def newfcn(x):
        return module.forward(x) + module.forward2(x) * 2

    assert isinstance(newfcn, PureFunction)
    assert len(newfcn.objparams()) == 2

    expr = lambda x, a, b: (x * a + b) + (x + a) * 2
    runtest_pfunc(newfcn, (x,), expr)

def runtest_pfunc(pfunc, params, expr):
    objparams = pfunc.objparams()

    # test original values
    res0 = pfunc(*params)
    res0_true = expr(*params, *objparams)
    assert torch.allclose(res0, res0_true)

    # test changing obj params
    objparams1 = [p + 1.0 for p in objparams]
    res1_true = expr(*params, *objparams1)
    with pfunc.useobjparams(objparams1):
        res1 = pfunc(*params)
    assert torch.allclose(res1, res1_true)

    # test recovery
    res2 = pfunc(*params)
    res2_true = res0_true
    assert torch.allclose(res2, res2_true)

    # test nested
    objparams3 = [p * 2.0 for p in objparams]
    res3_true = expr(*params, *objparams3)
    with pfunc.useobjparams(objparams1):
        assert torch.allclose(pfunc(*params), res1_true)
        with pfunc.useobjparams(objparams3):
            assert torch.allclose(pfunc(*params), res3_true)
        assert torch.allclose(pfunc(*params), res1_true)
    assert torch.allclose(pfunc(*params), res0_true)



import random
from collections import defaultdict
import torch
from torch.autograd import gradcheck, gradgradcheck
import xitorch as xt
import pytest
from xitorch.grad.jachess import hess
from xitorch.optimize import rootfinder, equilibrium, minimize
from xitorch._tests.utils import device_dtype_float_test

class DummyModule(xt.EditableModule):
    def __init__(self, A, addx=True, activation="sigmoid", sumoutput=False):
        super(DummyModule, self).__init__()
        self.A = A  # (nr, nr)
        self.addx = addx
        self.activation = {
            "sigmoid": lambda x: 1 / (1 + torch.exp(-x)),
            "cos": torch.cos,
            "square": lambda x: (x - 0.1) ** 2,
        }[activation]
        self.sumoutput = sumoutput
        self.biasdiff = True

    def set_diag_bias(self, diag, bias):
        self.diag = diag
        self.bias = bias
        self.biasdiff = bias.requires_grad

    def forward(self, x):
        # x: (nbatch, nr)
        # diag: (nbatch, nr)
        # bias: (nbatch, nr)
        nbatch, nr = x.shape
        x = x.unsqueeze(-1)
        A = self.A.unsqueeze(0).expand(nbatch, -1, -1)  # (nbatch, nr, nr)
        A = A + torch.diag_embed(self.diag)  # (nbatch, nr, nr)
        y = torch.bmm(A, x).squeeze(-1)  # (nbatch, nr)
        yr = self.activation(2 * y) + 2 * self.bias
        if self.addx:
            yr = yr + x.squeeze(-1)
        if self.sumoutput:
            yr = yr.sum()
        return yr

    def getparamnames(self, methodname, prefix=""):
        return [prefix + "A", prefix + "diag"] + ([prefix + "bias"] if self.biasdiff else [])

class DummyNNModule(torch.nn.Module):
    def __init__(self, A, addx=True, activation="sigmoid", sumoutput=False):
        super(DummyNNModule, self).__init__()
        self.A = A
        self.addx = addx
        self.activation = {
            "sigmoid": lambda x: 1 / (1 + torch.exp(-x)),
            "cos": torch.cos,
            "square": lambda x: (x - 0.1) ** 2,
        }[activation]
        self.sumoutput = sumoutput
        self.biasdiff = True

    def set_diag_bias(self, diag, bias):
        self.diag = diag
        self.bias = bias
        self.biasdiff = bias.requires_grad

    def forward(self, x):
        # x: (nbatch, nr)
        # diag: (nbatch, nr)
        # bias: (nbatch, nr)
        nbatch, nr = x.shape
        x = x.unsqueeze(-1)
        A = self.A.unsqueeze(0).expand(nbatch, -1, -1)  # (nbatch, nr, nr)
        A = A + torch.diag_embed(self.diag)  # (nbatch, nr, nr)
        y = torch.bmm(A, x).squeeze(-1)  # (nbatch, nr)
        yr = self.activation(2 * y) + 2 * self.bias
        if self.addx:
            yr = yr + x.squeeze(-1)
        if self.sumoutput:
            yr = yr.sum()
        return yr

class DummyNNInEditableModule(xt.EditableModule):
    # test to see if it still works if torch nn module in EditableModule

    def __init__(self, A, addx=True, activation="sigmoid", sumoutput=False):
        self.module = DummyNNModule(A, addx=addx, activation=activation, sumoutput=sumoutput)

    def set_diag_bias(self, diag, bias):
        self.module.set_diag_bias(diag, bias)

    def forward(self, x):
        return self.module.forward(x)

    def getparamnames(self, methodname, prefix=""):
        nnprefix = prefix + "module"  # torch nn prefix does not have no dot at the end while ours does
        return [name for (name, param) in self.module.named_parameters(prefix=nnprefix)]

class DummyModuleExplicit(xt.EditableModule):
    def __init__(self, addx=True):
        super(DummyModuleExplicit, self).__init__()
        self.addx = addx
        self.activation = lambda x: 1 / (1 + torch.exp(-x))

    def forward(self, x, A, diag, bias):
        nbatch, nr = x.shape
        x = x.unsqueeze(-1)
        A = A.unsqueeze(0).expand(nbatch, -1, -1)
        A = A + torch.diag_embed(diag)
        y = torch.bmm(A, x).squeeze(-1)
        yr = self.activation(2 * y) + 2 * bias
        if self.addx:
            yr = yr + x.squeeze(-1)
        return yr

    def getparamnames(self, methodname, prefix=""):
        return []

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "clss": [DummyModule, DummyNNModule, DummyNNInEditableModule],
})
def test_rootfinder(dtype, device, clss):
    torch.manual_seed(100)
    random.seed(100)

    nr = 2
    nbatch = 2
    fwd_options = {
        "method": "broyden1",
        "f_tol": 1e-9,
        "alpha": -0.5,
    }

    A    = torch.nn.Parameter((torch.randn((nr, nr)) * 0.5).to(dtype).requires_grad_())
    diag = torch.nn.Parameter(torch.randn((nbatch, nr)).to(dtype).requires_grad_())
    bias = torch.nn.Parameter(torch.zeros((nbatch, nr)).to(dtype).requires_grad_())
    y0 = torch.randn((nbatch, nr)).to(dtype)

    model = clss(A, addx=True)
    model.set_diag_bias(diag, bias)
    y = rootfinder(model.forward, y0, **fwd_options)
    f = model.forward(y)
    assert torch.allclose(f * 0, f)
    assert y.shape == y0.shape

    def getloss(A, y0, diag, bias):
        model = clss(A, addx=True)
        model.set_diag_bias(diag, bias)
        y = rootfinder(model.forward, y0, **fwd_options)
        return y

    # only check for real numbers or complex with DummyModule to save time
    checkgrad = not torch.is_complex(y0) or clss is DummyModule
    if checkgrad:
        gradcheck(getloss, (A, y0, diag, bias))
        gradgradcheck(getloss, (A, y0, diag, bias))

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "clss": [DummyModule, DummyNNModule],
    "method": ["broyden1", "anderson_acc"]
})
def test_equil(dtype, device, clss, method):
    torch.manual_seed(100)
    random.seed(100)

    nr = 2
    nbatch = 2
    if method == "broyden1":
        fwd_options = {
            "method": "broyden1",
            "f_tol": 1e-12,
            "alpha": -0.5,
        }
    else:
        fwd_options = {
            "method": "anderson_acc",
            "f_tol": 1e-12,
        }
    bck_options = {
        # NOTE: using "cg" fails the test, and using "gmres" produces an error
        # of re-entrant
        "method": "exactsolve",
    }

    A    = torch.nn.Parameter((torch.randn((nr, nr)) * 0.5).to(dtype).requires_grad_())
    diag = torch.nn.Parameter(torch.randn((nbatch, nr)).to(dtype).requires_grad_())
    bias = torch.nn.Parameter(torch.zeros((nbatch, nr)).to(dtype).requires_grad_())
    y0 = torch.randn((nbatch, nr)).to(dtype)

    model = clss(A, addx=False)
    model.set_diag_bias(diag, bias)
    y = equilibrium(model.forward, y0, bck_options=bck_options, **fwd_options)
    f = model.forward(y)
    assert torch.allclose(y, f)
    assert y.shape == y0.shape

    def getloss(A, y0, diag, bias):
        model = clss(A, addx=False)
        model.set_diag_bias(diag, bias)
        y = equilibrium(model.forward, y0, bck_options=bck_options, **fwd_options)
        return y

    # only check for real numbers or complex with DummyModule to save time
    checkgrad = not torch.is_complex(y0) or clss is DummyModule
    if checkgrad:
        gradcheck(getloss, (A, y0, diag, bias))
        gradgradcheck(getloss, (A, y0, diag, bias))

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "bias_is_tensor": [True, False]
})
def test_rootfinder_with_params(dtype, device, bias_is_tensor):
    torch.manual_seed(100)
    random.seed(100)

    nr = 2
    nbatch = 2
    fwd_options = {
        "method": "broyden1",
        "f_tol": 1e-9,
        "alpha": -0.5,
    }

    clss = DummyModuleExplicit
    A    = (torch.randn((nr, nr)) * 0.5).to(dtype).requires_grad_()
    diag = torch.randn((nbatch, nr)).to(dtype).requires_grad_()
    if bias_is_tensor:
        bias = torch.zeros((nbatch, nr)).to(dtype).requires_grad_()
    else:
        bias = 0.0
    y0 = torch.randn((nbatch, nr)).to(dtype)

    model = clss(addx=True)
    y = rootfinder(model.forward, y0, (A, diag, bias), **fwd_options)
    f = model.forward(y, A, diag, bias)
    assert torch.allclose(f * 0, f)

    def getloss(y0, A, diag, bias):
        model = clss(addx=True)
        y = rootfinder(model.forward, y0, (A, diag, bias), **fwd_options)
        return y

    # only check for real numbers or complex with bias_is_tensor to save time
    checkgrad = not torch.is_complex(y0) or bias_is_tensor
    if checkgrad:
        gradcheck(getloss, (y0, A, diag, bias))
        gradgradcheck(getloss, (y0, A, diag, bias))

@device_dtype_float_test(only64=True, additional_kwargs={
    "clss": [DummyModule, DummyNNModule],
    "method": ["broyden1", "gd"],
})
def test_minimize(dtype, device, clss, method):
    torch.manual_seed(400)
    random.seed(100)

    method_fwd_options = {
        "broyden1": {
            "max_niter": 50,
            "f_tol": 1e-9,
            "alpha": -0.5,
        },
        "gd": {
            "maxiter": 10000,
            "f_rtol": 1e-14,
            "x_rtol": 1e-14,
            "step": 2e-2,
        },
    }

    nr = 2
    nbatch = 2

    A    = torch.nn.Parameter((torch.randn((nr, nr)) * 0.5).to(dtype).requires_grad_())
    diag = torch.nn.Parameter(torch.randn((nbatch, nr)).to(dtype).requires_grad_())
    bias = torch.nn.Parameter(torch.zeros((nbatch, nr)).to(dtype).requires_grad_())
    y0 = torch.randn((nbatch, nr)).to(dtype)
    fwd_options = method_fwd_options[method]
    bck_options = {
        "rtol": 1e-9,
        "atol": 1e-9,
    }
    activation = "square"  # square activation makes it easy to optimize

    model = clss(A, addx=False, activation=activation, sumoutput=True)
    model.set_diag_bias(diag, bias)
    y = minimize(model.forward, y0, method=method, **fwd_options)

    # check the grad (must be close to 0)
    with torch.enable_grad():
        y1 = y.clone().requires_grad_()
        f = model.forward(y1)
    grady, = torch.autograd.grad(f, (y1,))
    assert torch.allclose(grady, grady * 0)

    # check the hessian (must be posdef)
    h = hess(model.forward, (y1,), idxs=0).fullmatrix()
    eigval = torch.linalg.eigvalsh(h)
    assert torch.all(eigval >= 0)

    def getloss(A, y0, diag, bias):
        model = clss(A, addx=False, activation=activation, sumoutput=True)
        model.set_diag_bias(diag, bias)
        y = minimize(model.forward, y0, method=method, bck_options=bck_options, **fwd_options)
        return y

    gradcheck(getloss, (A, y0, diag, bias))
    # pytorch 1.8's gradgradcheck fails if there are unrelated variables
    # I have made a PR to solve this and will be in 1.9
    gradgradcheck(getloss, (A, y0, diag, bias.detach()))

############## forward methods test ##############
@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "method": ["broyden1", "broyden2", "linearmixing"],
})
def test_rootfinder_methods(dtype, device, method):
    torch.manual_seed(100)
    random.seed(100)
    dtype = torch.float64

    nr = 3
    nbatch = 2
    default_fwd_options = {
        "f_tol": 1e-9,
        "alpha": -0.5,
    }
    # list the methods and the options here
    options = {
        "broyden1": default_fwd_options,
        "broyden2": default_fwd_options,
        "linearmixing": default_fwd_options,
    }[method]

    A    = torch.nn.Parameter((torch.randn((nr, nr)) * 0.5).to(dtype).requires_grad_())
    diag = torch.nn.Parameter(torch.randn((nbatch, nr)).to(dtype).requires_grad_())
    bias = torch.nn.Parameter(torch.zeros((nbatch, nr)).to(dtype).requires_grad_())
    y0 = torch.randn((nbatch, nr)).to(dtype)

    fwd_options = {**options, "method": method}
    model = DummyModule(A, addx=True)
    model.set_diag_bias(diag, bias)
    y = rootfinder(model.forward, y0, **fwd_options)
    f = model.forward(y)
    assert torch.allclose(f * 0, f)

@device_dtype_float_test(only64=True, include_complex=True, additional_kwargs={
    "method": ["broyden1", "broyden2", "linearmixing", "anderson_acc"],
})
def test_equil_methods(dtype, device, method):
    torch.manual_seed(100)
    random.seed(100)

    nr = 3
    nbatch = 2
    default_fwd_options = {
        "f_tol": 1e-9,
        "alpha": -0.5,
    }
    # list the methods and the options here
    options = {
        "broyden1": default_fwd_options,
        "broyden2": default_fwd_options,
        "linearmixing": default_fwd_options,
        "anderson_acc": {"f_tol": 1e-9},
    }[method]

    A    = torch.nn.Parameter((torch.randn((nr, nr)) * 0.5).to(dtype).requires_grad_())
    diag = torch.nn.Parameter(torch.randn((nbatch, nr)).to(dtype).requires_grad_())
    bias = torch.nn.Parameter(torch.zeros((nbatch, nr)).to(dtype).requires_grad_())
    y0 = torch.randn((nbatch, nr)).to(dtype)

    fwd_options = {**options, "method": method}
    model = DummyModule(A, addx=False)
    model.set_diag_bias(diag, bias)
    y = equilibrium(model.forward, y0, **fwd_options)
    f = model.forward(y)
    assert torch.allclose(y, f)

@device_dtype_float_test(only64=True, additional_kwargs={
    "method": ["broyden1", "broyden2", "linearmixing", "gd", "adam"]
})
def test_minimize_methods(dtype, device, method):
    torch.manual_seed(400)
    random.seed(100)

    nr = 3
    nbatch = 2
    default_fwd_options = {
        "max_niter": 50,
        "f_tol": 1e-9,
        "alpha": -1.0,
    }
    linearmixing_fwd_options = {
        "max_niter": 50,
        "f_tol": 3e-6,
        "alpha": -0.3,
    }
    gd_fwd_options = {
        "maxiter": 5000,
        "f_rtol": 1e-10,
        "x_rtol": 1e-10,
        "step": 1e-2,
    }
    # list the methods and the options here
    options = {
        "broyden1": default_fwd_options,
        "broyden2": default_fwd_options,
        "linearmixing": linearmixing_fwd_options,
        "gd": gd_fwd_options,
        "adam": gd_fwd_options,
    }[method]

    # specify higher atol for non-ideal method
    atol = defaultdict(lambda: 1e-8)
    atol["linearmixing"] = 3e-6

    A    = torch.nn.Parameter((torch.randn((nr, nr)) * 0.5).to(dtype).requires_grad_())
    diag = torch.nn.Parameter(torch.randn((nbatch, nr)).to(dtype).requires_grad_())
    # bias will be detached from the optimization line, so set it undifferentiable
    bias = torch.zeros((nbatch, nr)).to(dtype)
    y0 = torch.randn((nbatch, nr)).to(dtype)
    activation = "square"  # square activation makes it easy to optimize

    fwd_options = {**options, "method": method}
    model = DummyModule(A, addx=False, activation=activation, sumoutput=True)
    model.set_diag_bias(diag, bias)
    y = minimize(model.forward, y0, **fwd_options)

    # check the grad (must be close to 1)
    with torch.enable_grad():
        y1 = y.clone().requires_grad_()
        f = model.forward(y1)
    grady, = torch.autograd.grad(f, (y1,))
    assert torch.allclose(grady, grady * 0, atol=atol[method])

    # check the hessian (must be posdef)
    h = hess(model.forward, (y1,), idxs=0).fullmatrix()
    eigval = torch.linalg.eigvalsh(h)
    assert torch.all(eigval >= 0)

############## additional tests ##############
def test_min_not_stop_for_negative_value():
    # there was a bug where the minimizer stops right away if the value is negative
    # this test is written to make sure it does not happen again
    def fcn(a):
        return (a * a).sum() - 100.

    # the method must be non-rootfinder method
    with pytest.warns(UserWarning, match="converge"):
        method = "gd"
        a = torch.tensor(1.0, dtype=torch.float64)
        amin = minimize(fcn, a, method=method, step=0.2, f_rtol=0, x_rtol=0, verbose=True)
        amin_true = torch.zeros_like(amin)
    assert torch.allclose(amin, amin_true)

def test_minimizer_warnings():
    # test to see if it produces warnings
    def fcn(a):
        return (a * a).sum()

    with pytest.warns(UserWarning, match="converge"):
        # set it so that it will never converge
        a = torch.tensor(1.0, dtype=torch.float64)
        amin = minimize(fcn, a, method="gd", step=0.1, f_rtol=0, x_rtol=0, maxiter=10, verbose=True)



from __future__ import annotations
from typing import Sequence, Optional, List, Union
import warnings
import traceback
import torch
from abc import abstractmethod
from contextlib import contextmanager
from scipy.sparse.linalg import LinearOperator as spLinearOperator
from xitorch._core.editable_module import EditableModule
from xitorch.debug.modes import is_debug_enabled
from xitorch._utils.bcast import get_bcasted_dims

__all__ = ["LinearOperator"]

class LinearOperator(EditableModule):
    """
    ``LinearOperator`` is a base class designed to behave as a linear operator
    without explicitly determining the matrix.
    This ``LinearOperator`` should be able to operate as batched linear
    operators where its shape is ``(B1,B2,...,Bb,p,q)``
    with ``B*`` as the (optional) batch dimensions.

    For a user-defined class to behave as ``LinearOperator``, it must use
    ``LinearOperator`` as one of the parent and it has to have ``._mv()``
    method implemented and ``._getparamnames()`` if used in xitorch's
    functionals with torch grad enabled.
    """
    _implementation_checked    = False
    _is_mv_implemented         = False
    _is_mm_implemented         = False
    _is_rmv_implemented        = False
    _is_rmm_implemented        = False
    _is_fullmatrix_implemented = False
    _is_gpn_implemented        = False

    def __new__(cls, *args, **kwargs):
        # check the implemented functions in the class
        if not cls._implementation_checked:
            cls._is_mv_implemented = cls.__check_if_implemented("_mv")
            cls._is_mm_implemented = cls.__check_if_implemented("_mm")
            cls._is_rmv_implemented = cls.__check_if_implemented("_rmv")
            cls._is_rmm_implemented = cls.__check_if_implemented("_rmm")
            cls._is_fullmatrix_implemented = cls.__check_if_implemented("_fullmatrix")
            cls._is_gpn_implemented = cls.__check_if_implemented("_getparamnames")

            cls._implementation_checked = True

            if not cls._is_mv_implemented:
                raise RuntimeError("LinearOperator must have at least _mv(self) "
                                   "method implemented")
        return super(LinearOperator, cls).__new__(cls)

    @classmethod
    def __check_if_implemented(cls, methodname: str) -> bool:
        this_method = getattr(cls, methodname)
        base_method = getattr(LinearOperator, methodname)
        return this_method is not base_method

    @classmethod
    def m(cls, mat: torch.Tensor, is_hermitian: Optional[bool] = None):
        """
        Class method to wrap a matrix into ``LinearOperator``.

        Arguments
        ---------
        mat: torch.Tensor
            Matrix to be wrapped in the ``LinearOperator``.
        is_hermitian: bool or None
            Indicating if the matrix is Hermitian. If ``None``, the symmetry
            will be checked. If supplied as a bool, there is no check performed.

        Returns
        -------
        LinearOperator
            Linear operator object that represents the matrix.

        Example
        -------
        .. testsetup:: *

            import torch
            import xitorch
            torch.manual_seed(100)

        .. doctest::

            >>> mat = torch.rand(1,3,1,2)  # 1x2 matrix with (1,3) batch dimensions
            >>> linop = xitorch.LinearOperator.m(mat)
            >>> print(linop)
            MatrixLinearOperator with shape (1, 3, 1, 2):
               tensor([[[[0.1117, 0.8158]],
            <BLANKLINE>
                        [[0.2626, 0.4839]],
            <BLANKLINE>
                        [[0.6765, 0.7539]]]])
        """
        if is_hermitian is None:
            if mat.shape[-2] != mat.shape[-1]:
                is_hermitian = False
            else:
                is_hermitian = torch.allclose(mat, mat.transpose(-2, -1).conj())
        elif is_hermitian:
            # check the hermitian
            if not torch.allclose(mat, mat.transpose(-2, -1).conj()):
                raise RuntimeError("The linear operator is indicated to be hermitian, but the matrix is not")

        return MatrixLinearOperator(mat, is_hermitian)

    def __init__(self, shape: Sequence[int],
                 is_hermitian: bool = False,
                 dtype: Optional[torch.dtype] = None,
                 device: Optional[torch.device] = None,
                 _suppress_hermit_warning: bool = False) -> None:

        super(LinearOperator, self).__init__()
        if len(shape) < 2:
            raise RuntimeError("The shape must have at least 2 dimensions")
        self._shape = shape
        self._batchshape = list(shape[:-2])
        self._is_hermitian = is_hermitian
        self._dtype = dtype if dtype is not None else torch.float32
        self._device = device if device is not None else torch.device("cpu")
        if is_hermitian and shape[-1] != shape[-2]:
            raise RuntimeError("The object is indicated as Hermitian, but the shape is not square")

        # check which methods are implemented
        if not _suppress_hermit_warning and self._is_hermitian and \
           (self._is_rmv_implemented or self._is_rmm_implemented):
            warnings.warn("The LinearOperator is Hermitian with implemented "
                          "rmv or rmm. We will use the mv and mm methods "
                          "instead",
                          stacklevel=2)

    def __repr__(self) -> str:
        return "LinearOperator (%s) with shape %s, dtype = %s, device = %s" % \
            (self.__class__.__name__, _shape2str(self.shape), self.dtype, self.device)

    @abstractmethod
    def _getparamnames(self, prefix: str = "") -> List[str]:
        """
        List the self's parameters that affecting the ``LinearOperator``.
        This is for the derivative purpose.

        Arguments
        ---------
        prefix: str
            The prefix to be appended in front of the parameters name.
            This usually contains the dots.

        Returns
        -------
        list of str
            List of parameter names (including the prefix) that affecting
            the ``LinearOperator``.
        """
        return []

    @abstractmethod
    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Abstract method to be implemented for matrix-vector multiplication.
        Required for all ``LinearOperator`` objects.
        """
        pass

    # @abstractmethod
    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Abstract method to be implemented for transposed matrix-vector multiplication.
        Optional. If not implemented, it will use the adjoint trick to compute ``.rmv()``.
        Usually implemented for efficiency reasons.
        """
        raise NotImplementedError()

    # @abstractmethod # (optional)
    def _mm(self, x: torch.Tensor) -> torch.Tensor:
        """
        Abstract method to be implemented for matrix-matrix multiplication.
        If not implemented, then it uses batched version of matrix-vector
        multiplication.
        Usually this is implemented for efficiency reasons.
        """
        raise NotImplementedError()

    # @abstractmethod
    def _rmm(self, x: torch.Tensor) -> torch.Tensor:
        """
        Abstract method to be implemented for transposed matrix-matrix multiplication.
        If not implemented, then it uses batched version of transposed matrix-vector
        multiplication.
        Usually this is implemented for efficiency reasons.
        """
        raise NotImplementedError()

    # @abstractmethod
    def _fullmatrix(self) -> torch.Tensor:
        raise NotImplementedError()

    # linear operators must have a set of parameters that affects most of
    # the methods (i.e. mm, mv, rmm, rmv)
    def getlinopparams(self) -> Sequence[torch.Tensor]:
        return self.getuniqueparams("mm")

    @contextmanager
    def uselinopparams(self, *params):
        methodname = "mm"
        try:
            _orig_params_ = self.getuniqueparams(methodname)
            self.setuniqueparams(methodname, *params)
            yield self
        finally:
            self.setuniqueparams(methodname, *_orig_params_)

    ############# implemented functions ################
    def mv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix-vector operation to vector ``x`` with shape ``(...,q)``.
        The batch dimensions of ``x`` need not be the same as the batch dimensions
        of the ``LinearOperator``, but it must be broadcastable.

        Arguments
        ---------
        x: torch.tensor
            The vector with shape ``(...,q)`` where the linear operation is operated on

        Returns
        -------
        y: torch.tensor
            The result of the linear operation with shape ``(...,p)``
        """
        self.__assert_if_init_executed()
        if x.shape[-1] != self.shape[-1]:
            raise RuntimeError("Cannot operate .mv on shape %s. Expected (...,%d)" %
                               (str(tuple(x.shape)), self.shape[-1]))

        return self._mv(x)

    def mm(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix-matrix operation to matrix ``x`` with shape ``(...,q,r)``.
        The batch dimensions of ``x`` need not be the same as the batch dimensions
        of the ``LinearOperator``, but it must be broadcastable.

        Arguments
        ---------
        x: torch.tensor
            The matrix with shape ``(...,q,r)`` where the linear operation is operated on.

        Returns
        -------
        y: torch.tensor
            The result of the linear operation with shape ``(...,p,r)``
        """
        self.__assert_if_init_executed()
        if x.shape[-2] != self.shape[-1]:
            raise RuntimeError("Cannot operate .mm on shape %s. Expected (...,%d,*)" %
                               (str(tuple(x.shape)), self.shape[-1]))

        xbatchshape = list(x.shape[:-2])
        if self._is_mm_implemented:
            return self._mm(x)
        else:
            # use batched mv as mm

            # move the last dimension to the very first dimension to be broadcasted
            if len(xbatchshape) < len(self._batchshape):
                xbatchshape = [1] * (len(self._batchshape) - len(xbatchshape)) + xbatchshape
            x1 = x.reshape(1, *xbatchshape, *x.shape[-2:])
            xnew = x1.transpose(0, -1).squeeze(-1)  # (r,...,q)

            # apply batched mv and restore the initial shape
            ynew = self._mv(xnew)  # (r,...,p)
            y = ynew.unsqueeze(-1).transpose(0, -1).squeeze(0)  # (...,p,r)
            return y

    def rmv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix-vector adjoint operation to vector ``x`` with shape ``(...,p)``,
        i.e. ``A^H x``.
        The batch dimensions of ``x`` need not be the same as the batch dimensions
        of the ``LinearOperator``, but it must be broadcastable.

        Arguments
        ---------
        x: torch.tensor
            The vector of shape ``(...,p)`` where the adjoint linear operation is operated at.

        Returns
        -------
        y: torch.tensor
            The result of the adjoint linear operation with shape ``(...,q)``
        """
        self.__assert_if_init_executed()
        if x.shape[-1] != self.shape[-2]:
            raise RuntimeError("Cannot operate .rmv on shape %s. Expected (...,%d)" %
                               (str(tuple(x.shape)), self.shape[-2]))

        if self._is_hermitian:
            return self._mv(x)
        elif not self._is_rmv_implemented:
            return self.__adjoint_rmv(x)
        return self._rmv(x)

    def rmm(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix-matrix adjoint operation to matrix ``x`` with shape ``(...,p,r)``,
        i.e. ``A^H X``.
        The batch dimensions of ``x`` need not be the same as the batch dimensions
        of the ``LinearOperator``, but it must be broadcastable.

        Arguments
        ---------
        x: torch.Tensor
            The matrix of shape ``(...,p,r)`` where the adjoint linear operation is operated on.

        Returns
        -------
        y: torch.Tensor
            The result of the adjoint linear operation with shape ``(...,q,r)``.
        """
        self.__assert_if_init_executed()
        if x.shape[-2] != self.shape[-2]:
            raise RuntimeError("Cannot operate .rmm on shape %s. Expected (...,%d,*)" %
                               (str(tuple(x.shape)), self.shape[-2]))

        if self._is_hermitian:
            return self.mm(x)

        xbatchshape = list(x.shape[:-2])
        if self._is_rmm_implemented:
            return self._rmm(x)
        else:
            # use batched mv as mm
            rmv = self._rmv if self._is_rmv_implemented else self.rmv

            # move the last dimension to the very first dimension to be broadcasted
            if len(xbatchshape) < len(self._batchshape):
                xbatchshape = [1] * (len(self._batchshape) - len(xbatchshape)) + xbatchshape
            x1 = x.reshape(1, *xbatchshape, *x.shape[-2:])  # (1,...,p,r)
            xnew = x1.transpose(0, -1).squeeze(-1)  # (r,...,p)

            # apply batched mv and restore the initial shape
            ynew = rmv(xnew)  # (r,...,q)
            y = ynew.unsqueeze(-1).transpose(0, -1).squeeze(0)  # (...,q,r)
            return y

    def fullmatrix(self) -> torch.Tensor:
        if self._is_fullmatrix_implemented:
            return self._fullmatrix()
        else:
            self.__assert_if_init_executed()
            nq = self._shape[-1]
            V = torch.eye(nq, dtype=self._dtype, device=self._device)  # (nq,nq)
            return self.mm(V)  # (B1,B2,...,Bb,np,nq)

    def scipy_linalg_op(self):
        to_tensor = lambda x: torch.tensor(x, dtype=self.dtype, device=self.device)
        return spLinearOperator(
            shape=self.shape,
            matvec=lambda v: self.mv(to_tensor(v)).detach().cpu().numpy(),
            rmatvec=lambda v: self.rmv(to_tensor(v)).detach().cpu().numpy(),
            matmat=lambda v: self.mm(to_tensor(v)).detach().cpu().numpy(),
            rmatmat=lambda v: self.rmm(to_tensor(v)).detach().cpu().numpy(),
        )

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """"""
        # just to remove the docstring from EditableModule because user
        # does not need to know about this function

        if methodname in ["mv", "rmv", "mm", "rmm", "fullmatrix"]:
            return self._getparamnames(prefix=prefix)
        else:
            raise KeyError("getparamnames for method %s is not implemented" % methodname)

    ############# cached properties ################
    @property
    def H(self):
        """
        Returns a LinearOperator representing the Hermite / transposed of the
        self LinearOperator.

        Returns
        -------
        LinearOperator
            The Hermite / transposed LinearOperator
        """
        if self._is_hermitian:
            return self
        elif isinstance(self, MatrixLinearOperator):
            return LinearOperator.m(self.fullmatrix().transpose(-2, -1).conj())
        return AdjointLinearOperator(self)

    ############# special functions ################
    def matmul(self, b: LinearOperator, is_hermitian: bool = False):
        """
        Returns a LinearOperator representing `self @ b`.

        Arguments
        ---------
        b: LinearOperator
            Other linear operator
        is_hermitian: bool
            Flag to indicate if the resulting LinearOperator is Hermitian.

        Returns
        -------
        LinearOperator
            LinearOperator representing `self @ b`
        """
        # returns linear operator that represents self @ b
        if self.shape[-1] != b.shape[-2]:
            raise RuntimeError("Mismatch shape of matmul operation: %s and %s" % (self.shape, b.shape))
        if isinstance(self, MatrixLinearOperator) and isinstance(b, MatrixLinearOperator):
            return LinearOperator.m(self.fullmatrix() @ b.fullmatrix(), is_hermitian=is_hermitian)
        return MatmulLinearOperator(self, b, is_hermitian=is_hermitian)

    def __add__(self, b: LinearOperator):
        assert isinstance(b, LinearOperator), \
            "Only addition with another LinearOperator is supported"
        if self.shape[-2:] != b.shape[-2:]:
            raise RuntimeError("Mismatch shape of add operation: %s and %s" % (self.shape, b.shape))
        if isinstance(self, MatrixLinearOperator) and isinstance(b, MatrixLinearOperator):
            return LinearOperator.m(self.fullmatrix() + b.fullmatrix())
        return AddLinearOperator(self, b)

    def __sub__(self, b: LinearOperator):
        assert isinstance(b, LinearOperator), \
            "Only subtraction with another LinearOperator is supported"
        if self.shape[-2:] != b.shape[-2:]:
            raise RuntimeError("Mismatch shape of add operation: %s and %s" % (self.shape, b.shape))
        if isinstance(self, MatrixLinearOperator) and isinstance(b, MatrixLinearOperator):
            return LinearOperator.m(self.fullmatrix() - b.fullmatrix())
        return AddLinearOperator(self, b, -1)

    def __rsub__(self, b: LinearOperator):
        return b.__sub__(self)

    def __mul__(self, f: Union[int, float]):
        if not (isinstance(f, int) or isinstance(f, float)):
            raise TypeError("LinearOperator multiplication only supports integer or floating point")
        if isinstance(self, MatrixLinearOperator):
            return LinearOperator.m(self.fullmatrix() * f)
        return MulLinearOperator(self, f)

    def __rmul__(self, f: Union[int, float]):
        return self.__mul__(f)

    ############# properties ################
    @property
    def dtype(self) -> torch.dtype:
        return self._dtype

    @property
    def device(self) -> torch.device:
        return self._device

    @property
    def shape(self) -> Sequence[int]:
        return self._shape

    @property
    def is_hermitian(self) -> bool:
        return self._is_hermitian

    # implementation
    @property
    def is_mv_implemented(self) -> bool:
        return True

    @property
    def is_mm_implemented(self) -> bool:
        return self._is_mm_implemented

    @property
    def is_rmv_implemented(self) -> bool:
        return self._is_rmv_implemented

    @property
    def is_rmm_implemented(self) -> bool:
        return self._is_rmm_implemented

    @property
    def is_fullmatrix_implemented(self) -> bool:
        return self._is_fullmatrix_implemented

    @property
    def is_getparamnames_implemented(self) -> bool:
        return self._is_gpn_implemented

    ############ debug functions ##############
    def check(self, warn: Optional[bool] = None) -> None:
        """
        Perform checks to make sure the ``LinearOperator`` behaves as a proper
        linear operator.

        Arguments
        ---------
        warn: bool or None
            If ``True``, then raises a warning to the user that the check might slow
            down the program. This is to remind the user to turn off the check
            when not in a debugging mode.
            If ``None``, it will raise a warning if it runs not in a debug mode, but
            will be silent if it runs in a debug mode.

        Raises
        ------
        RuntimeError
            Raised if an error is raised when performing linear operations of the
            object (e.g. calling ``.mv()``, ``.mm()``, etc)
        AssertionError
            Raised if the linear operations do not behave as proper linear operations.
            (e.g. not scaling linearly)
        """
        if warn is None:
            warn = not is_debug_enabled()
        if warn:
            msg = "The linear operator check is performed. This might slow down your program."
            warnings.warn(msg, stacklevel=2)
        checklinop(self)
        print("Check linear operator done")

    ############ private functions #################
    def __adjoint_rmv(self, xt: torch.Tensor) -> torch.Tensor:
        # xt: (*BY, p)
        # xdummy: (*BY, q)
        # calculate the right matvec multiplication by using the adjoint trick

        BY = xt.shape[:-1]
        BA = self.shape[:-2]
        BAY = get_bcasted_dims(BY, BA)

        # calculate y = Ax
        p, q = self.shape[-2:]
        xdummy = torch.zeros((*BAY, q), dtype=xt.dtype, device=xt.device).requires_grad_()
        with torch.enable_grad():
            y = self.mv(xdummy)  # (*BAY, p)

        # calculate (dL/dx)^T = A^T (dL/dy)^T with (dL/dy)^T = xt
        xt2 = xt.contiguous().expand_as(y)  # (*BAY, p)
        res = torch.autograd.grad(y, xdummy, grad_outputs=xt2,
                                  create_graph=torch.is_grad_enabled())[0]  # (*BAY, q)
        return res

    # def __check_if_implemented(self, methodname: str) -> bool:
    #     this_method = getattr(self, methodname).__func__
    #     base_method = getattr(LinearOperator, methodname)
    #     return this_method is not base_method

    def __assert_if_init_executed(self):
        if not hasattr(self, "_shape"):
            raise RuntimeError("super().__init__ must be executed first")

############## special linear operators ##############
class AdjointLinearOperator(LinearOperator):
    def __init__(self, obj: LinearOperator):
        super(AdjointLinearOperator, self).__init__(
            shape=(*obj.shape[:-2], obj.shape[-1], obj.shape[-2]),
            is_hermitian=obj.is_hermitian,
            dtype=obj.dtype,
            device=obj.device,
            _suppress_hermit_warning=True,
        )
        self.obj = obj

    def __repr__(self):
        return "AdjointLinearOperator with shape %s of:\n - %s" % \
            (_shape2str(self.shape), _indent(self.obj.__repr__(), 3))

    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        if not self.obj.is_rmv_implemented:
            raise RuntimeError("The ._rmv of must be implemented to call .H.mv()")
        return self.obj._rmv(x)

    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        return self.obj._mv(x)

    def _getparamnames(self, prefix: str = "") -> List[str]:
        return self.obj._getparamnames(prefix=prefix + "obj.")

    @property
    def H(self):
        return self.obj

class MatmulLinearOperator(LinearOperator):
    def __init__(self, a: LinearOperator, b: LinearOperator, is_hermitian: bool = False):
        shape = (*get_bcasted_dims(a.shape[:-2], b.shape[:-2]), a.shape[-2], b.shape[-1])
        super(MatmulLinearOperator, self).__init__(
            shape=shape,
            is_hermitian=is_hermitian,
            dtype=a.dtype,
            device=a.device,
            _suppress_hermit_warning=True,
        )
        self.a = a
        self.b = b

    def __repr__(self):
        return "MatmulLinearOperator with shape %s of:\n * %s\n * %s" % \
            (_shape2str(self.shape),
             _indent(self.a.__repr__(), 3),
             _indent(self.b.__repr__(), 3))

    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        return self.a._mv(self.b._mv(x))

    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        return self.b.rmv(self.a.rmv(x))

    def _getparamnames(self, prefix: str = "") -> List[str]:
        return self.a._getparamnames(prefix=prefix + "a.") + \
            self.b._getparamnames(prefix=prefix + "b.")

class AddLinearOperator(LinearOperator):
    def __init__(self, a: LinearOperator, b: LinearOperator, mul: int = 1):
        shape = (*get_bcasted_dims(a.shape[:-2], b.shape[:-2]), a.shape[-2], b.shape[-1])
        is_hermitian = a.is_hermitian and b.is_hermitian
        super(AddLinearOperator, self).__init__(
            shape=shape,
            is_hermitian=is_hermitian,
            dtype=a.dtype,
            device=a.device,
            _suppress_hermit_warning=True,
        )
        self.a = a
        self.b = b
        assert mul == 1 or mul == -1
        self.mul = mul

    def __repr__(self):
        return "AddLinearOperator with shape %s of:\n * %s\n * %s" % \
            (_shape2str(self.shape),
             _indent(self.a.__repr__(), 3),
             _indent(self.b.__repr__(), 3))

    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        return self.a._mv(x) + self.mul * self.b._mv(x)

    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        return self.a.rmv(x) + self.mul * self.b.rmv(x)

    def _getparamnames(self, prefix: str = "") -> List[str]:
        return self.a._getparamnames(prefix=prefix + "a.") + \
            self.b._getparamnames(prefix=prefix + "b.")

class MulLinearOperator(LinearOperator):
    def __init__(self, a: LinearOperator, f: Union[int, float]):
        shape = a.shape
        is_hermitian = a.is_hermitian
        super(MulLinearOperator, self).__init__(
            shape=shape,
            is_hermitian=is_hermitian,
            dtype=a.dtype,
            device=a.device,
            _suppress_hermit_warning=True,
        )
        self.a = a
        self.f = f

    def __repr__(self):
        return "MulLinearOperator with shape %s of: \n * %s\n * %s" % \
            (_shape2str(self.shape),
             _indent(self.a.__repr__(), 3),
             _indent(self.f.__repr__(), 3))

    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        return self.a._mv(x) * self.f

    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        return self.a._rmv(x) * self.f

    def _getparamnames(self, prefix: str = "") -> List[str]:
        pnames = self.a._getparamnames(prefix=prefix + "a.")
        return pnames

class MatrixLinearOperator(LinearOperator):
    def __init__(self, mat: torch.Tensor, is_hermitian: bool) -> None:

        super(MatrixLinearOperator, self).__init__(
            shape=mat.shape,
            is_hermitian=is_hermitian,
            dtype=mat.dtype,
            device=mat.device,
            _suppress_hermit_warning=True,
        )
        self.mat = mat

    def __repr__(self):
        return "MatrixLinearOperator with shape %s:\n   %s" % \
            (_shape2str(self.shape), _indent(self.mat.__repr__(), 3))

    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)

    def _mm(self, x: torch.Tensor) -> torch.Tensor:
        return torch.matmul(self.mat, x)

    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        return torch.matmul(self.mat.transpose(-2, -1).conj(), x.unsqueeze(-1)).squeeze(-1)

    def _rmm(self, x: torch.Tensor) -> torch.Tensor:
        return torch.matmul(self.mat.transpose(-2, -1).conj(), x)

    def _fullmatrix(self) -> torch.Tensor:
        return self.mat

    def _getparamnames(self, prefix: str = "") -> List[str]:
        return [prefix + "mat"]

def checklinop(linop: LinearOperator) -> None:
    """
    Check if the implemented mv and mm can receive the possible shapes and returns
    the correct shape. If an error is found, then this function raise AssertionError.

    Argument
    --------
    * linop: LinearOperator instance
        The instance of LinearOperator to be checked

    Exception
    ---------
    * AssertionError
        Raised if there is a shape mismatch
    * RuntimeError
        Raised if there is an error when evaluating the .mv, .mm, .rmv, or .rmm methods
    """
    shape = linop.shape
    p, q = shape[-2:]
    batchshape = shape[:-2]

    def runtest(methodname, xshape, yshape):
        x = torch.rand(xshape, dtype=linop.dtype, device=linop.device)
        fcn = getattr(linop, methodname)
        try:
            y = fcn(x)
        except Exception:
            s = traceback.format_exc()
            msg = "An error is raised from .%s with input shape: %s (linear operator shape: %s)\n" % \
                (methodname, tuple(xshape), tuple(linop.shape))
            msg += "--- full traceback ---\n%s" % s
            raise RuntimeError(msg)
        msg = "The output shape of .%s is not correct. Input: %s, expected output: %s, output: %s" % \
            (methodname, tuple(x.shape), tuple(yshape), tuple(y.shape))
        msg += "\n" + str(linop)
        assert list(y.shape) == list(yshape), msg

        # linearity test
        x2 = 1.25 * x
        y2 = fcn(x2)
        msg = "Linearity check fails\n%s\n" % str(linop)
        assert torch.allclose(y2, 1.25 * y), msg
        y0 = fcn(0 * x)
        assert torch.allclose(y0, y * 0), "Linearity check (with 0) fails\n" + str(linop)

        # batched test
        xnew = torch.cat((x.unsqueeze(0), x2.unsqueeze(0)), dim=0)
        ynew = fcn(xnew)  # (2, ..., q)
        msg = "Batched test fails (expanding batches changes the results)" + str(linop)
        assert torch.allclose(ynew[0], y), msg
        assert torch.allclose(ynew[1], y2), msg

    # generate shapes
    mv_xshapes = [
        (q,),
        (1, q),
        (1, 1, q),
        (*batchshape, q),
        (1, *batchshape, q),
    ]
    mv_yshapes = [
        (*batchshape, p),
        (*batchshape, p) if len(batchshape) >= 1 else (1, p),
        (*batchshape, p) if len(batchshape) >= 2 else (1, 1, p),
        (*batchshape, p),
        (1, *batchshape, p)
    ]
    # test matvec and matmat, run input in multiple shapes to make sure no error is raised
    r = 2
    for (mv_xshape, mv_yshape) in zip(mv_xshapes, mv_yshapes):
        runtest("mv", mv_xshape, mv_yshape)
        runtest("mm", (*mv_xshape, r), (*mv_yshape, r))

    if not linop.is_rmv_implemented:
        return

    rmv_xshapes = [
        (p,),
        (1, p),
        (1, 1, p),
        (*batchshape, p),
        (1, *batchshape, p),
    ]
    rmv_yshapes = [
        (*batchshape, q),
        (*batchshape, q) if len(batchshape) >= 1 else (1, q),
        (*batchshape, q) if len(batchshape) >= 2 else (1, 1, q),
        (*batchshape, q),
        (1, *batchshape, q)
    ]
    for (rmv_xshape, rmv_yshape) in zip(rmv_xshapes, rmv_yshapes):
        runtest("rmv", rmv_xshape, rmv_yshape)
        runtest("rmm", (*rmv_xshape, r), (*rmv_yshape, r))

########### repr helper functions ###########
def _indent(s, nspace):
    # give indentation of the second line and next lines
    spaces = " " * nspace
    lines = [spaces + c if i > 0 else c for i, c in enumerate(s.split("\n"))]
    return "\n".join(lines)

def _shape2str(shape):
    return "(%s)" % (", ".join([str(s) for s in shape]))



import torch

"""
This file contains functions for some linear algebra and basic operations of
torch.tensor.
"""

def tallqr(V, MV=None):
    # faster QR for tall and skinny matrix
    # V: (*BV, na, nguess)
    # MV: (*BM, na, nguess) where M is the basis to make Q M-orthogonal
    # if MV is None, then MV=V
    if MV is None:
        MV = V
    VTV = torch.matmul(V.transpose(-2, -1), MV)  # (*BMV, nguess, nguess)
    R = torch.linalg.cholesky(VTV.transpose(-2, -1).conj()).transpose(-2, -1).conj()  # (*BMV, nguess, nguess)
    Rinv = torch.inverse(R)  # (*BMV, nguess, nguess)
    Q = torch.matmul(V, Rinv)
    return Q, R

def to_fortran_order(V):
    # V: (...,nrow,ncol)
    # outV: (...,nrow,ncol)

    # check if it is in C-contiguous
    if V.is_contiguous():
        # return V.set_(V.storage(), V.storage_offset(), V.size(), tuple(reversed(V.stride())))
        return V.transpose(-2, -1).contiguous().transpose(-2, -1)
    elif V.transpose(-2, -1).is_contiguous():
        return V
    else:
        raise RuntimeError("Only the last two dimensions can be made Fortran order.")

def convert_none_grads_to_zeros(grads, inputs):
    is_tuple = isinstance(grads, tuple)
    if is_tuple:
        grads = list(grads)
    for i in range(len(grads)):
        g = grads[i]
        if g is None:
            grads[i] = torch.zeros_like(inputs[i])
    if is_tuple:
        grads = tuple(grads)
    return grads

def create_random_square_matrix(
        n, is_hermitian=False,
        min_eival=1.0, max_eival=1.0, minabs_eival=0.0, seed=-1):

    dtype = torch.float64
    eivals = torch.linspace(min_eival, max_eival, n, dtype=dtype)

    # clip absolute eivals to minabs_eival
    idx = eivals.abs() < minabs_eival
    eivals[idx] = torch.sign(eivals[idx]) * minabs_eival

    eivals = torch.diag_embed(eivals)
    if seed > 0:
        torch.manual_seed(seed)
    if is_hermitian:
        eivecs = create_random_ortho_matrix(n, seed=seed)
        mat = torch.matmul(torch.matmul(eivecs.transpose(-2, -1), eivals), eivecs)
        mat = (mat + mat.transpose(-2, -1)) * 0.5
        return mat
    else:
        a = torch.randn((n, n), dtype=dtype)
        a = a / a.norm(dim=-2, keepdim=True)
        return torch.matmul(torch.matmul(a.inverse(), eivals), a)

def create_random_ortho_matrix(n, seed=-1):
    dtype = torch.float64
    if seed > 0:
        torch.manual_seed(seed)
    a = torch.randn((n, n), dtype=dtype)
    q, r = torch.linalg.qr(a)
    return q



import torch
from abc import abstractmethod
from xitorch.grad import jac

# taking most of the part from SciPy

__all__ = ["BroydenFirst", "BroydenSecond", "LinearMixing"]

class Jacobian(object):
    """
    Base class for the Jacobians used in rootfinder algorithms.
    """
    @abstractmethod
    def setup(self, x0, y0, func):
        pass

    @abstractmethod
    def solve(self, v, tol=0):
        pass

    @abstractmethod
    def update(self, x, y):
        pass

class BroydenFirst(Jacobian):
    """
    Approximating the Jacobian based on Broyden's first approximation.

    [1] B.A. van der Rotten, PhD thesis,
        "A limited memory Broyden method to solve high-dimensional
        systems of nonlinear equations". Mathematisch Instituut,
        Universiteit Leiden, The Netherlands (2003).
    """

    def __init__(self, alpha=None, uv0=None, max_rank=None):
        # The initial guess of inverse Jacobian is `-alpha * I + u v^T`.
        # `max_rank` indicates the maximum rank of the Jacoabian before
        # reducing it
        self.alpha = alpha
        self.uv0 = uv0
        self.max_rank = max_rank

    def setup(self, x0, y0, func):
        self.x_prev = x0
        self.y_prev = y0

        if self.max_rank is None:
            self.max_rank = float('inf')

        if self.alpha is None:
            normy0 = torch.norm(y0)
            ones = torch.ones_like(normy0)
            if normy0:
                self.alpha = 0.5 * torch.max(torch.norm(x0), ones) / normy0
            else:
                self.alpha = ones

        if self.uv0 == "svd":
            self.uv0 = _get_svd_uv0(func, x0)

        # setup the approximate inverse Jacobian
        self.Gm = LowRankMatrix(-self.alpha, self.uv0, "restart")
        # do not uncomment the line below! it causes memory leak
        # I left it here as a lesson for me and us to never repeat this mistake
        # self._reduce = lambda: self.Gm.reduce(self.max_rank)

    def _reduce(self):
        # reduce the size of Gm
        # initially it was a lambda function, but it causes a leak, so
        # I arranged into a method to remove the leak
        self.Gm.reduce(self.max_rank)

    def solve(self, v, tol=0):
        res = self.Gm.mv(v)
        return res

    def update(self, x, y):
        dy = y - self.y_prev
        dx = x - self.x_prev
        # update Gm
        self._update(x, y, dx, dy, dx.norm(), dy.norm())

        self.y_prev = y
        self.x_prev = x

    def _update(self, x, y, dx, dy, dxnorm, dynorm):
        # keep the rank small
        self._reduce()

        v = self.Gm.rmv(dx)
        c = dx - self.Gm.mv(dy)
        d = v / torch.dot(dy, v)
        self.Gm = self.Gm.append(c, d)

class BroydenSecond(BroydenFirst):
    """
    Inverse Jacobian approximation based on Broyden's second method.

    [1] B.A. van der Rotten, PhD thesis,
        "A limited memory Broyden method to solve high-dimensional
        systems of nonlinear equations". Mathematisch Instituut,
        Universiteit Leiden, The Netherlands (2003).
    """

    def _update(self, x, y, dx, dy, dxnorm, dynorm):
        # keep the rank small
        self._reduce()

        v = dy
        c = dx - self.Gm.mv(dy)
        d = v / (dynorm * dynorm)
        self.Gm = self.Gm.append(c, d)

class LinearMixing(Jacobian):
    def __init__(self, alpha=None):
        # The initial guess of inverse Jacobian is ``-alpha * I``
        if alpha is None:
            alpha = -1.0
        self.alpha = alpha

    def setup(self, x0, y0, func):
        pass

    def solve(self, v, tol=0):
        return -v * self.alpha

    def update(self, x, y):
        pass

class LowRankMatrix(object):
    # represents a matrix of `\alpha * I + \sum_n c_n d_n^T`
    def __init__(self, alpha, uv0, reduce_method):
        self.alpha = alpha
        if uv0 is None:
            self.cns = []
            self.dns = []
        else:
            cn0, dn0 = uv0
            self.cns = [cn0]
            self.dns = [dn0]
        self.reduce_method = {
            "restart": 0,
            "simple": 1
        }[reduce_method]

    def mv(self, v):
        res = self.alpha * v
        for i in range(len(self.dns)):
            res += self.cns[i] * torch.dot(self.dns[i], v)
        return res

    def rmv(self, v):
        res = self.alpha * v
        for i in range(len(self.dns)):
            res += self.dns[i] * torch.dot(self.cns[i], v)
        return res

    def append(self, c, d):
        self.cns.append(c)
        self.dns.append(d)
        if len(self.cns) >= torch.numel(c):
            return FullRankMatrix(self.alpha, self.cns, self.dns)
        return self

    def reduce(self, max_rank, **otherparams):
        if len(self.cns) > max_rank:
            if self.reduce_method == 0:  # restart
                del self.cns[:]
                del self.dns[:]
            elif self.reduce_method == 1:  # simple
                n = len(self.cns)
                del self.cns[:n - max_rank]
                del self.dns[:n - max_rank]

class FullRankMatrix(object):
    def __init__(self, alpha, cns, dns):
        size = torch.numel(cns[0])
        dtype, device = cns[0].dtype, cns[0].device
        self.mat = torch.eye(size, dtype=dtype, device=device)
        self.mat *= alpha
        for i in range(len(cns)):
            self.mat += torch.ger(cns[i], dns[i])

    def mv(self, v):
        res = torch.matmul(self.mat, v)
        return res

    def rmv(self, v):
        return torch.matmul(self.mat.T, v)

    def append(self, c, d):
        self.mat += torch.ger(c, d)
        return self

    def reduce(self, max_rank, **kwargs):
        pass  # ???

def _get_svd_uv0(func, x0):
    from xitorch.linalg import svd
    # raise RuntimeError
    fjac = jac(func, (x0.clone().requires_grad_(),), idxs=[0])[0]
    # u: (n, 1), s: (1,), vh: (1, n)
    u, s, vh = svd(fjac, k=1, mode="lowest", method="davidson", min_eps=1e-3)
    sinv_sqrt = 1. / torch.sqrt(torch.clamp(s, min=0.1))
    uv0 = (sinv_sqrt * vh.squeeze(-2), sinv_sqrt * u.squeeze(-1))
    return uv0



import inspect
from typing import Union, Any, Mapping, Sequence, Optional, Callable, List, \
    ItemsView, Generator, Tuple

def get_methods_docstr(
        cls_or_func: Callable,
        methods: Union[Sequence[Callable], Mapping[str, Any]],
        ignore_kwargs: Optional[List[str]] = None) -> str:
    """
    Get the full docstring of a class or a function. Full docstring is the
    main docstring of the class or function plus docstrings of the methods
    (usually to describe the options available specifically for a specific
    method).

    Arguments
    ---------
    * cls_or_func: class or callable
        Main class or function
    * methods: list or dict
        If a list, it lists the methods with the same method name options.
        If a dict, it contains the method name as the keys and the method
        function or class as the value.
    * ignore_kwargs: list or None
        List of string on which kwargs to be ignored in addition to ["params"]

    Returns
    -------
    * full_docstr: str
        The full docstring of cls_or_func
    """
    method_template = """
    Methods
    -------
    method="{name}"

        .. code-block:: python

            {mainname}(..., {kwargs_sig})
    """
    res = cls_or_func.__doc__ or ""
    mainname = cls_or_func.__name__

    def_ignore_kwargs = ["params"]
    if ignore_kwargs is None:
        ignore_kwargs = []
    ignore_kwargs = ignore_kwargs + def_ignore_kwargs

    if isinstance(methods, dict):
        generator: Union[ItemsView[str, Any], Generator[Tuple[str, Any], None, None]] = methods.items()
    elif isinstance(methods, list):
        generator = ((method.__name__, method) for method in methods)
    else:
        raise TypeError("methods must be a list or a dict")

    for name, method in generator:
        # get the signatures
        sigparams = inspect.signature(method).parameters
        kwargs_sig_list = ['method="%s"' % name]
        kwargs_sig_list2 = ["%s=%s" % (pname, val) for pname, val in _get_default_parameters(sigparams, ignore_kwargs)]
        kwargs_sig_list = kwargs_sig_list + (["*"] if len(kwargs_sig_list2) > 0 else []) + kwargs_sig_list2
        kwargs_sig = ", ".join(kwargs_sig_list)

        # add the method name
        methoddoc = method.__doc__
        res = res + method_template.format(
            mainname=mainname,
            name=name,
            kwargs_sig=kwargs_sig,
        )
        if methoddoc is not None:
            method_doc = _add_indent(method.__doc__, " " * 4)
            res = res + method_doc
    return res

def _get_default_parameters(parameters, ignore_kwargs: Sequence[str]):
    empty = inspect.Parameter.empty
    for paramname in parameters:
        if paramname in ignore_kwargs:
            continue
        defval = parameters[paramname].default
        if defval == empty:
            continue
        defval = defval if not isinstance(defval, str) else '"%s"' % defval
        yield paramname, defval

def _add_indent(s: str, indent: str) -> str:
    return "\n".join([indent + line for line in s.split("\n")])
